from copy import deepcopy
from haproxy.line import Line

import pytest


DEFAULT_DATA = {
    'syslog_date': 'Dec  9 13:01:26',
    'process_name_and_pid': 'localhost haproxy[28029]:',
    'client_ip': '127.0.0.1',
    'client_port': 2345,
    'accept_date': '09/Dec/2013:12:59:46.633',
    'frontend_name': 'loadbalancer',
    'backend_name': 'default',
    'server_name': 'instance8',
    'tq': 0,
    'tw': 51536,
    'tc': 1,
    'tr': 48082,
    'tt': '99627',
    'status': '200',
    'bytes': '83285',
    'act': '87',
    'fe': '89',
    'be': '98',
    'srv': '1',
    'retries': '20',
    'queue_server': 2,
    'queue_backend': 67,
    'headers': ' {77.24.148.74}',
    'http_request': 'GET /path/to/image HTTP/1.1',
}


class LinesGenerator:
    def __init__(self, line_format):
        self.data = deepcopy(DEFAULT_DATA)
        self.line_format = line_format

    def __call__(self, *args, **kwargs):
        self.data.update(**kwargs)
        self.data['client_ip_and_port'] = '{client_ip}:{client_port}'.format(
            **self.data
        )
        self.data[
            'server_names'
        ] = '{frontend_name} {backend_name}/{server_name}'.format(**self.data)
        self.data['timers'] = '{tq}/{tw}/{tc}/{tr}/{tt}'.format(**self.data)
        self.data['status_and_bytes'] = '{status} {bytes}'.format(**self.data)
        self.data['connections_and_retries'] = '{act}/{fe}/{be}/{srv}/{retries}'.format(
            **self.data
        )
        self.data['queues'] = '{queue_server}/{queue_backend}'.format(**self.data)

        log_line = self.line_format.format(**self.data)
        return Line(log_line)


@pytest.fixture()
def default_line_data():
    return DEFAULT_DATA


@pytest.fixture()
def line_factory():
    # queues and headers parameters are together because if no headers are
    # saved the field is completely empty and thus there is no double space
    # between queue backend and http request.
    raw_line = (
        '{syslog_date} {process_name_and_pid} {client_ip_and_port} '
        '[{accept_date}] {server_names} {timers} {status_and_bytes} '
        '- - ---- {connections_and_retries} {queues}{headers} '
        '"{http_request}"'
    )
    generator = LinesGenerator(raw_line)
    return generator



from datetime import datetime
from datetime import timedelta
from haproxy import commands

import pytest


def check_output(cmd, output, expected, capsys):
    """Validate the output of commands."""
    name = cmd.command_line_name().upper()
    cmd.results(output=output)
    output_text = capsys.readouterr().out
    if output == 'json':
        assert f'{{"{name}": {expected}}}' in output_text
    else:
        assert f'{name}\n====' in output_text
        assert f'====\n{expected}\n' in output_text


@pytest.mark.parametrize(
    ('klass', 'expected'),
    [
        (commands.StatusCodesCounter, 'status_codes_counter'),
        (commands.AverageResponseTime, 'average_response_time'),
        (commands.Counter, 'counter'),
        (commands.IpCounter, 'ip_counter'),
    ],
)
def test_commands_names(klass, expected):
    """Check that the command line name of command classes are generated correctly."""
    assert klass.command_line_name() == expected


def test_counter_results():
    """Test the Counter command.

    It plain and simply counts all the lines passed to it.
    """
    cmd = commands.Counter()
    assert cmd.raw_results() == 0
    for x in range(3):
        cmd(x)

    assert cmd.raw_results() == 3


@pytest.mark.parametrize('output', [None, 'json'])
def test_counter_output(capsys, output):
    """Test the Counter command.

    It plain and simply counts all the lines passed to it.
    """
    cmd = commands.Counter()
    for x in range(3):
        cmd(x)
    check_output(cmd, output, 3, capsys)


def test_http_methods_results(line_factory):
    """Test the HTTPMethods command.

    It creates a breakdown of how many times each HTTP verb has been used.
    """
    cmd = commands.HttpMethods()
    assert cmd.raw_results() == {}
    for verb, count in (('POST', 4), ('GET', 3), ('PUT', 2)):
        line = line_factory(http_request=f'{verb} /path/to/image HTTP/1.1')
        for _ in range(count):
            cmd(line)
    results = cmd.raw_results()
    assert len(results) == 3
    assert results['POST'] == 4
    assert results['GET'] == 3
    assert results['PUT'] == 2


@pytest.mark.parametrize(
    ('output', 'expected'),
    [(None, '- PUT: 2\n- GET: 1'), ('json', '[{"PUT": 2}, {"GET": 1}]')],
)
def test_http_methods_output(line_factory, capsys, output, expected):
    """Test the HTTPMethods command.

    It creates a breakdown of how many times each HTTP verb has been used.
    """
    cmd = commands.HttpMethods()
    for verb, count in (('GET', 1), ('PUT', 2)):
        line = line_factory(http_request=f'{verb} /path/to/image HTTP/1.1')
        for _ in range(count):
            cmd(line)
    check_output(cmd, output, expected, capsys)


def test_ip_counter_results(line_factory):
    """Test the IpCounter command.

    It creates a breakdown of how many times each IP has been used.
    """
    cmd = commands.IpCounter()
    assert cmd.raw_results() == {}
    for ip, count in (('192.168.0.1', 4), ('172.4.3.2', 3), ('8.7.6.5', 2)):
        line = line_factory(headers=f' {{{ip}}}')
        for _ in range(count):
            cmd(line)
    results = cmd.raw_results()
    assert len(results) == 3
    assert results['192.168.0.1'] == 4
    assert results['172.4.3.2'] == 3
    assert results['8.7.6.5'] == 2


@pytest.mark.parametrize(
    ('output', 'expected'),
    [
        (None, '- 172.4.3.2: 3\n- 8.7.6.5: 2'),
        ('json', '[{"172.4.3.2": 3}, {"8.7.6.5": 2}]'),
    ],
)
def test_ip_counter_output(line_factory, capsys, output, expected):
    """Test the IpCounter command.

    It creates a breakdown of how many times each IP has been used.
    """
    cmd = commands.IpCounter()
    for ip, count in (('172.4.3.2', 3), ('8.7.6.5', 2)):
        line = line_factory(headers=f' {{{ip}}}')
        for _ in range(count):
            cmd(line)
    check_output(cmd, output, expected, capsys)


def test_top_ips_results(line_factory):
    """Test the TopIps command.

    It lists the 10 most used IPs, and how much where they used.
    """
    cmd = commands.TopIps()
    assert cmd.raw_results() == []
    for ip, count in ((f'192.168.0.{x}', x) for x in range(11)):
        line = line_factory(headers=f' {{{ip}}}')
        for _ in range(count):
            cmd(line)
    results = cmd.raw_results()
    assert len(results) == 10
    assert results[0] == ('192.168.0.10', 10)
    assert results[1] == ('192.168.0.9', 9)
    assert results[2] == ('192.168.0.8', 8)
    assert results[3] == ('192.168.0.7', 7)
    assert results[4] == ('192.168.0.6', 6)
    assert results[5] == ('192.168.0.5', 5)
    assert results[6] == ('192.168.0.4', 4)
    assert results[7] == ('192.168.0.3', 3)
    assert results[8] == ('192.168.0.2', 2)
    assert results[9] == ('192.168.0.1', 1)


def test_top_ips_print_results(line_factory):
    """Test the TopIps command.

    Ensure that when they are printed, only 10 results are shown.
    """
    cmd = commands.TopIps()
    for ip, count in ((f'192.168.0.{x}', x) for x in range(14)):
        line = line_factory(headers=f' {{{ip}}}')
        for _ in range(count):
            cmd(line)
    results = cmd.print_data()
    results = [x for x in results.split('\n') if x]
    assert len(results) == 10
    assert results[0] == '- 192.168.0.13: 13'
    assert results[-1] == '- 192.168.0.4: 4'


@pytest.mark.parametrize(
    ('output', 'expected'),
    [
        (None, '- 192.168.0.2: 2\n- 192.168.0.1: 1'),
        ('json', '[{"192.168.0.2": 2}, {"192.168.0.1": 1}]'),
    ],
)
def test_top_ips_output(line_factory, capsys, output, expected):
    """Test the TopIps command.

    It lists the 10 most used IPs, and how much where they used.
    """
    cmd = commands.TopIps()
    assert cmd.raw_results() == []
    for ip, count in ((f'192.168.0.{x}', x) for x in range(3)):
        line = line_factory(headers=f' {{{ip}}}')
        for _ in range(count):
            cmd(line)
    check_output(cmd, output, expected, capsys)


def test_status_codes_counter_results(line_factory):
    """Test the StatusCodesCounter command.

    It creates a breakdown of which status codes have been used and how many each.
    """
    cmd = commands.StatusCodesCounter()
    assert cmd.raw_results() == {}
    for status_code, count in (('200', 4), ('301', 3), ('500', 2)):
        line = line_factory(status=status_code)
        for _ in range(count):
            cmd(line)
    results = cmd.raw_results()
    assert len(results) == 3
    assert results['200'] == 4
    assert results['301'] == 3
    assert results['500'] == 2


@pytest.mark.parametrize(
    ('output', 'expected'),
    [(None, '- 301: 3\n- 500: 2'), ('json', '[{"301": 3}, {"500": 2}]')],
)
def test_status_codes_counter_output(line_factory, capsys, output, expected):
    """Test the StatusCodesCounter command.

    It creates a breakdown of which status codes have been used and how many each.
    """
    cmd = commands.StatusCodesCounter()
    for status_code, count in (('301', 3), ('500', 2)):
        line = line_factory(status=status_code)
        for _ in range(count):
            cmd(line)
    check_output(cmd, output, expected, capsys)


def test_request_path_counter_results(line_factory):
    """Test the RequestPathCounter command.

    It creates a breakdown of how many times each URL path has been used.
    """
    cmd = commands.RequestPathCounter()
    assert cmd.raw_results() == {}
    for path, count in (('/image/one', 4), ('/video/two', 3), ('/article/three', 2)):
        line = line_factory(http_request=f'GET {path} HTTP/1.1')
        for _ in range(count):
            cmd(line)
    results = cmd.raw_results()
    assert len(results) == 3
    assert results['/image/one'] == 4
    assert results['/video/two'] == 3
    assert results['/article/three'] == 2


@pytest.mark.parametrize(
    ('output', 'expected'),
    [
        (None, '- /video/two: 3\n- /article/three: 2'),
        ('json', '[{"/video/two": 3}, {"/article/three": 2}]'),
    ],
)
def test_request_path_counter_output(line_factory, capsys, output, expected):
    """Test the RequestPathCounter command.

    It creates a breakdown of how many times each URL path has been used.
    """
    cmd = commands.RequestPathCounter()
    for path, count in (('/video/two', 3), ('/article/three', 2)):
        line = line_factory(http_request=f'GET {path} HTTP/1.1')
        for _ in range(count):
            cmd(line)
    check_output(cmd, output, expected, capsys)


def test_slow_requests_results(line_factory):
    """Test the SlowRequests command.

    It lists all requests that took more than 1000 milliseconds to respond.
    """
    cmd = commands.SlowRequests()
    assert cmd.raw_results() == []
    for total_time in (1003, 987, 456, 2013, 45000, 1000, 3200, 999):
        cmd(line_factory(tr=total_time))
    results = cmd.raw_results()
    assert results == [1000, 1003, 2013, 3200, 45000]


@pytest.mark.parametrize(
    ('output', 'expected'),
    [
        (None, [1000, 1003, 2013, 3200, 45000]),
        ('json', '[1000, 1003, 2013, 3200, 45000]'),
    ],
)
def test_slow_requests_output(line_factory, capsys, output, expected):
    """Test the SlowRequests command.

    It lists all requests that took more than 1000 milliseconds to respond.
    """
    cmd = commands.SlowRequests()
    for total_time in (1003, 987, 456, 2013, 45000, 1000, 3200, 999):
        cmd(line_factory(tr=total_time))
    check_output(cmd, output, expected, capsys)


def test_top_request_paths_results(line_factory):
    """Test the TopRequestPaths command.

    It lists the 10 most used URL paths, and how much where they used.
    """
    cmd = commands.TopRequestPaths()
    assert cmd.raw_results() == []
    for path, count in ((f'/file/{x}', x) for x in range(11)):
        line = line_factory(http_request=f'GET {path} HTTP/1.1')
        for _ in range(count):
            cmd(line)
    results = cmd.raw_results()
    assert len(results) == 10
    assert results[0] == ('/file/10', 10)
    assert results[1] == ('/file/9', 9)
    assert results[2] == ('/file/8', 8)
    assert results[3] == ('/file/7', 7)
    assert results[4] == ('/file/6', 6)
    assert results[5] == ('/file/5', 5)
    assert results[6] == ('/file/4', 4)
    assert results[7] == ('/file/3', 3)
    assert results[8] == ('/file/2', 2)
    assert results[9] == ('/file/1', 1)


def test_top_request_paths_print_results(line_factory):
    """Test the TopRequestPaths command.

    Ensure that when they are printed, only 10 results are shown.
    """
    cmd = commands.TopRequestPaths()
    for path, count in ((f'/file/{x}', x) for x in range(14)):
        line = line_factory(http_request=f'GET {path} HTTP/1.1')
        for _ in range(count):
            cmd(line)
    results = cmd.print_data()
    results = [x for x in results.split('\n') if x]
    assert len(results) == 10
    assert results[0] == '- /file/13: 13'
    assert results[-1] == '- /file/4: 4'


@pytest.mark.parametrize(
    ('output', 'expected'),
    [
        (None, '- /file/2: 2\n- /file/1: 1'),
        ('json', '[{"/file/2": 2}, {"/file/1": 1}]'),
    ],
)
def test_top_request_paths_output(line_factory, capsys, output, expected):
    """Test the TopRequestPaths command.

    It lists the 10 most used URL paths, and how much where they used.
    """
    cmd = commands.TopRequestPaths()
    for path, count in ((f'/file/{x}', x) for x in range(3)):
        line = line_factory(http_request=f'GET {path} HTTP/1.1')
        for _ in range(count):
            cmd(line)
    check_output(cmd, output, expected, capsys)


def test_slow_requests_counter_results(line_factory):
    """Test the SlowRequestsCounter command.

    It counts how many requests took more than 1000 milliseconds to complete.
    """
    cmd = commands.SlowRequestsCounter()
    assert cmd.raw_results() == 0
    for total_time in (1003, 987, 456, 2013, 45000, 1000, 3200, 999):
        cmd(line_factory(tr=total_time))
    results = cmd.raw_results()
    assert results == 5


@pytest.mark.parametrize('output', [None, 'json'])
def test_slow_requests_counter_output(line_factory, capsys, output):
    """Test the SlowRequestsCounter command.

    It counts how many requests took more than 1000 milliseconds to complete.
    """
    cmd = commands.SlowRequestsCounter()
    for total_time in (1003, 987, 456, 2013, 45000, 1000, 3200, 999):
        cmd(line_factory(tr=total_time))
    check_output(cmd, output, 5, capsys)


@pytest.mark.parametrize(
    ('series', 'average'),
    [
        ((1003, 987, 456, 2013, 1000, 3200, 999), 1379.71),
        ((110, -1, 110), 110),  # aborted connections are ignored
        ((45, 30, 0), 25),  # responses that take 0 milliseconds are still counted
    ],
)
def test_average_response_time_results(line_factory, series, average):
    """Test the AverageResponseTime command.

    Returns the average response time of all valid requests.
    """
    cmd = commands.AverageResponseTime()
    assert cmd.raw_results() == 0.0
    for total_time in series:
        cmd(line_factory(tr=total_time))
    results = cmd.raw_results()
    assert results == average


@pytest.mark.parametrize('output', [None, 'json'])
def test_average_response_time_output(line_factory, capsys, output):
    """Test the AverageResponseTime command.

    Returns the average response time of all valid requests.
    """
    cmd = commands.AverageResponseTime()
    for total_time in (
        40,
        30,
    ):
        cmd(line_factory(tr=total_time))
    check_output(cmd, output, 35.0, capsys)


@pytest.mark.parametrize(
    ('series', 'average'),
    [
        ((1003, 987, 456, 2013, 1000, 3200, 999), 1379.71),
        ((110, -1, 110), 110),  # aborted connections are ignored
        ((45, 30, 0), 25),  # requests that do not wait at all are still counted
    ],
)
def test_average_waiting_time_results(line_factory, series, average):
    """Test the AverageWaitingTime command.

    Returns the average time requests had to wait to get processed.
    """
    cmd = commands.AverageWaitingTime()
    assert cmd.raw_results() == 0.0
    for wait_time in series:
        cmd(line_factory(tw=wait_time))
    results = cmd.raw_results()
    assert results == average


@pytest.mark.parametrize('output', [None, 'json'])
def test_average_waiting_time_output(line_factory, capsys, output):
    """Test the AverageWaitingTime command.

    Returns the average time requests had to wait to get processed.
    """
    cmd = commands.AverageWaitingTime()
    for wait_time in (40, 30):
        cmd(line_factory(tw=wait_time))
    check_output(cmd, output, 35.0, capsys)


def test_server_load_results(line_factory):
    """Test the ServerLoad command.

    It creates a breakdown of how many requests each server processed.
    """
    cmd = commands.ServerLoad()
    assert cmd.raw_results() == {}
    for name, count in (('server4', 4), ('server3', 3), ('server5', 5)):
        line = line_factory(server_name=name)
        for _ in range(count):
            cmd(line)
    results = cmd.raw_results()
    assert len(results) == 3
    assert results['server5'] == 5
    assert results['server4'] == 4
    assert results['server3'] == 3


@pytest.mark.parametrize(
    ('output', 'expected'),
    [
        (None, '- server5: 5\n- server3: 3'),
        ('json', '[{"server5": 5}, {"server3": 3}]'),
    ],
)
def test_server_load_output(line_factory, capsys, output, expected):
    """Test the ServerLoad command.

    It creates a breakdown of how many requests each server processed.
    """
    cmd = commands.ServerLoad()
    for name, count in (('server3', 3), ('server5', 5)):
        line = line_factory(server_name=name)
        for _ in range(count):
            cmd(line)
    check_output(cmd, output, expected, capsys)


def test_queue_peaks_no_lines_results(line_factory):
    """Test the QueuePeaks command.

    If there are no log lines processed, nothing should be returned.
    """
    cmd = commands.QueuePeaks()
    assert cmd.raw_results() == []


def test_queue_peaks_no_queues(line_factory):
    """Test the QueuePeaks command.

    If there are no log lines processed, nothing should be returned.
    """
    cmd = commands.QueuePeaks()
    now = datetime.now()
    for second in range(4):
        accept_date = now.replace(second=second).strftime('%d/%b/%Y:%H:%M:%S.%f')
        cmd(line_factory(queue_backend=0, accept_date=accept_date))
    assert len(cmd.requests) == 4
    assert cmd.raw_results() == []


@pytest.mark.parametrize(
    ('date', 'expected_key'),
    [
        ('10/Dec/2019:15:40:12.12345', 1575988812.12345),
        ('15/Jan/2017:05:23:05.456', 1484454185.456),
        ('15/Jan/2017:05:23:05.0', 1484454185.0),
    ],
)
def test_queue_peaks_generated_keys(line_factory, date, expected_key):
    """Test the QueuePeaks command.

    Check how the keys for the requests dictionary are generated.
    """
    cmd = commands.QueuePeaks()
    cmd(line_factory(queue_backend=0, accept_date=date))
    keys = list(cmd.requests.keys())
    # account for a 1h difference, if UTC is used (as in CI)
    assert expected_key - 4000 <= keys[0] <= expected_key + 4000
    # check that microseconds are exact though
    assert expected_key - int(expected_key) == keys[0] - int(keys[0])


def test_queue_peaks_details(line_factory):
    """Test the QueuePeaks command.

    Check the information returned for each peak.
    """
    cmd = commands.QueuePeaks()
    for microseconds, queue in enumerate([0, 4, 7, 8, 19, 4, 0]):
        line = line_factory(
            queue_backend=queue, accept_date=f'15/Jan/2017:05:23:05.{microseconds}'
        )
        cmd(line)
    day = datetime(year=2017, month=1, day=15, hour=5, minute=23, second=5)
    results = cmd.raw_results()
    assert len(results) == 1
    peak_info = results[0]
    assert peak_info['peak'] == 19
    assert peak_info['span'] == 5
    assert peak_info['started'] == day.replace(microsecond=100000)
    assert peak_info['finished'] == day.replace(microsecond=600000)


def test_queue_peaks_multiple_sorted(line_factory):
    """Test the QueuePeaks command.

    Peaks information are returned sorted by date.
    """
    cmd = commands.QueuePeaks()
    for microseconds, queue in enumerate([0, 4, 0, 0, 19, 4, 0]):
        line = line_factory(
            queue_backend=queue, accept_date=f'15/Jan/2017:05:23:05.{microseconds}'
        )
        cmd(line)
    day = datetime(year=2017, month=1, day=15, hour=5, minute=23, second=5)
    results = cmd.raw_results()
    assert len(results) == 2
    assert results[0]['peak'] == 4
    assert results[0]['started'] == day.replace(microsecond=100000)
    assert results[1]['peak'] == 19
    assert results[1]['started'] == day.replace(microsecond=400000)


def test_queue_peaks_already_started(line_factory):
    """Test the QueuePeaks command.

    Check that QueuePeaks handles the corner case of a peak that has already started.
    """
    cmd = commands.QueuePeaks()
    for microseconds, queue in enumerate([4, 19, 0]):
        line = line_factory(
            queue_backend=queue, accept_date=f'15/Jan/2017:05:23:05.{microseconds}'
        )
        cmd(line)
    day = datetime(year=2017, month=1, day=15, hour=5, minute=23, second=5)
    results = cmd.raw_results()
    assert len(results) == 1
    peak_info = results[0]
    assert peak_info['peak'] == 19
    assert peak_info['span'] == 2
    assert peak_info['started'] == day
    assert peak_info['finished'] == day.replace(microsecond=200000)


def test_queue_peaks_did_not_finish(line_factory):
    """Test the QueuePeaks command.

    Check that QueuePeaks handles the corner case of a peak that does not finish.
    """
    cmd = commands.QueuePeaks()
    for microseconds, queue in enumerate([4, 19, 12]):
        line = line_factory(
            queue_backend=queue, accept_date=f'15/Jan/2017:05:23:05.{microseconds}'
        )
        cmd(line)
    day = datetime(year=2017, month=1, day=15, hour=5, minute=23, second=5)
    results = cmd.raw_results()
    assert len(results) == 1
    peak_info = results[0]
    assert peak_info['peak'] == 19
    assert peak_info['span'] == 3
    assert peak_info['started'] == day
    assert peak_info['finished'] == day.replace(microsecond=200000)


@pytest.mark.parametrize(
    ('output', 'expected'),
    [
        (
            None,
            '- peak: 4 - span: 1 - started: 2017-01-15T05:23:05.100000 - finished: 2017-01-15T05:23:05.200000\n'
            '- peak: 19 - span: 2 - started: 2017-01-15T05:23:05.400000 - finished: 2017-01-15T05:23:05.600000',
        ),
        (
            'json',
            '[{"peak": 4, "span": 1, "started": "2017-01-15T05:23:05.100000", "finished": "2017-01-15T05:23:05.200000"}, '
            '{"peak": 19, "span": 2, "started": "2017-01-15T05:23:05.400000", "finished": "2017-01-15T05:23:05.600000"}]',
        ),
    ],
)
def test_queue_peaks_output(line_factory, capsys, output, expected):
    """Test the QueuePeaks command.

    Peaks information are returned sorted by date.
    """
    cmd = commands.QueuePeaks()
    for microseconds, queue in enumerate([0, 4, 0, 0, 19, 4, 0]):
        line = line_factory(
            queue_backend=queue, accept_date=f'15/Jan/2017:05:23:05.{microseconds}'
        )
        cmd(line)
    check_output(cmd, output, expected, capsys)


def test_connection_type_results(line_factory):
    """Test the ConnectionType command.

    It counts how many requests have been made by SSL, and which ones not.
    """
    cmd = commands.ConnectionType()
    assert cmd.raw_results() == (0, 0)
    for path, count in (('/Virtual:443/something', 4), ('/something', 2)):
        line = line_factory(http_request=f'GET {path} HTTP/1.1')
        for _ in range(count):
            cmd(line)
    assert cmd.raw_results() == (4, 2)


@pytest.mark.parametrize(
    ('output', 'expected'),
    [(None, '- https: 4\n- http: 2'), ('json', '[{"https": 4}, {"http": 2}]')],
)
def test_connection_type_output(line_factory, capsys, output, expected):
    """Test the ConnectionType command.

    It counts how many requests have been made by SSL, and which ones not.
    """
    cmd = commands.ConnectionType()
    for path, count in (('/Virtual:443/something', 4), ('/something', 2)):
        line = line_factory(http_request=f'GET {path} HTTP/1.1')
        for _ in range(count):
            cmd(line)
    check_output(cmd, output, expected, capsys)


def test_requests_per_minute_results(line_factory):
    """Test the RequestsPerMinute command.

    It counts how many requests have been made per minute.
    """
    cmd = commands.RequestsPerMinute()
    assert cmd.raw_results() == []
    now = datetime.now()
    # to avoid leaping into the next/previous minute with the timedeltas below
    now = now.replace(second=30)
    microseconds = timedelta(microseconds=200)
    seconds = timedelta(seconds=5)
    minutes = timedelta(minutes=5)
    hours = timedelta(hours=2)
    dates = [
        now,
        now + microseconds,
        now - microseconds,
        now + seconds,
        now - seconds,
        now + minutes,
        now - minutes,
        now + hours,
        now - hours,
    ]
    for time in dates:
        cmd(line_factory(accept_date=f'{time:%d/%b/%Y:%H:%M:%S.%f}'))
    results = cmd.raw_results()
    assert len(results) == 5
    assert results[0][1] == 1
    assert results[1][1] == 1
    assert results[2][1] == 5  # now and the +- microseconds and +- seconds
    assert results[3][1] == 1
    assert results[4][1] == 1


@pytest.mark.parametrize('output', [None, 'json'])
def test_requests_per_minute_output(line_factory, capsys, output):
    """Test the RequestsPerMinute command.

    It counts how many requests have been made per minute.
    """
    cmd = commands.RequestsPerMinute()
    now = datetime.now()
    for time in (now, now + timedelta(hours=2)):
        cmd(line_factory(accept_date=f'{time:%d/%b/%Y:%H:%M:%S.%f}'))
    name = cmd.command_line_name().upper()
    cmd.results(output=output)
    output_text = capsys.readouterr().out
    if output == 'json':
        assert f'{{"{name}": ' in output_text
        # this is quite fuzzy to not have to fiddle with the date formatting
        # change it once we hit 2030 :)
        assert ':00": 1}, {"202' in output_text
    else:
        assert f'{name}\n====' in output_text
        # this is quite fuzzy to not have to fiddle with the date formatting
        assert ':00: 1\n- ' in output_text


def test_requests_per_hour_results(line_factory):
    """Test the RequestsPerHour command.

    It counts how many requests have been made per hour.
    """
    cmd = commands.RequestsPerHour()
    assert cmd.raw_results() == []
    specific_date = datetime(year=2022, month=12, day=3, hour=14, minute=10, second=30)
    minutes = timedelta(minutes=5)
    hours = timedelta(hours=2)
    dates = [
        specific_date,
        specific_date + minutes,
        specific_date - minutes,
        specific_date + hours,
        specific_date - hours,
        specific_date + hours * 2,
        specific_date - hours * 2,
    ]
    for time in dates:
        cmd(line_factory(accept_date=f'{time:%d/%b/%Y:%H:%M:%S.%f}'))
    results = cmd.raw_results()
    assert len(results) == 5
    assert results[0][1] == 1
    assert results[1][1] == 1
    assert results[2][1] == 3  # now and the +- minutes
    assert results[3][1] == 1
    assert results[4][1] == 1


@pytest.mark.parametrize('output', [None, 'json'])
def test_requests_per_hour_output(line_factory, capsys, output):
    """Test the RequestsPerHour command.

    It counts how many requests have been made per hour.
    """
    cmd = commands.RequestsPerHour()
    now = datetime.now()
    for time in (now, now + timedelta(hours=2)):
        cmd(line_factory(accept_date=f'{time:%d/%b/%Y:%H:%M:%S.%f}'))
    name = cmd.command_line_name().upper()
    cmd.results(output=output)
    output_text = capsys.readouterr().out
    if output == 'json':
        assert f'{{"{name}": ' in output_text
        # this is quite fuzzy to not have to fiddle with the date formatting
        # change it once we hit 2030 :)
        assert ':00": 1}, {"202' in output_text
    else:
        assert f'{name}\n====' in output_text
        # this is quite fuzzy to not have to fiddle with the date formatting
        assert ':00: 1\n- ' in output_text


def test_print_results_and_output(line_factory, capsys):
    """Test the Print command.

    It simply prints the verbatim line.
    """
    cmd = commands.Print()
    assert cmd.raw_results() is None
    for path in ('/first-thing-to-do', '/second/thing/to-do'):
        cmd(line_factory(http_request=f'GET {path} HTTP/1.1'))
    assert cmd.raw_results() is None
    output_text = capsys.readouterr().out
    lines = output_text.split('\n')
    assert len(lines) == 3
    assert '/first-thing-to-do' in lines[0]
    assert '/second/thing/to-do' in lines[1]
    assert lines[2] == ''



from haproxy import filters

import pytest


@pytest.mark.parametrize(
    ('to_filter', 'to_check', 'result'),
    [
        ('1.2.3.4', '1.2.3.4', True),
        ('2.3.4.5', '5.3.5.4', False),
        ('2001:db8::8a2e:370:7334', '2001:db8::8a2e:370:7334', True),
        ('2001:db8::8a2e:370:7334', '2001:db8::8a2e:456:7321', False),
    ],
)
def test_filter_ip(line_factory, to_filter, to_check, result):
    """Check that filter_ip filter works as expected."""
    current_filter = filters.filter_ip(to_filter)
    headers = f' {{{to_check}}}'
    line = line_factory(headers=headers)
    assert current_filter(line) is result


@pytest.mark.parametrize(
    ('to_filter', 'to_check', 'result'),
    [
        ('1.2.3', '1.2.3.4', True),
        ('1.2.3', '1.2.3.78', True),
        ('2.3.4.5', '5.3.5.4', False),
        ('2001:db8', '2001:db8::8a2e:370:7334', True),
        ('2001:db8', '2001:db8::8a2e:456:7321', True),
        ('2134:db8', '2001:db8::8a2e:456:7321', False),
    ],
)
def test_filter_ip_range(line_factory, to_filter, to_check, result):
    """Check that filter_ip_range filter works as expected."""
    current_filter = filters.filter_ip_range(to_filter)
    headers = f' {{{to_check}}}'
    line = line_factory(headers=headers)
    assert current_filter(line) is result


@pytest.mark.parametrize(
    ('path', 'result'),
    [
        ('/path/to/image', True),
        ('/something/else', False),
        ('/another/image/here', True),
    ],
)
def test_filter_path(line_factory, path, result):
    """Check that filter_path filter works as expected."""
    current_filter = filters.filter_path('/image')
    http_request = f'GET {path} HTTP/1.1'
    line = line_factory(http_request=http_request)
    assert current_filter(line) is result


@pytest.mark.parametrize(
    ('path', 'result'),
    [
        ('/ssl_path:443/image', True),
        ('/something/else', False),
        ('/another:443/ssl', True),
    ],
)
def test_filter_ssl(line_factory, path, result):
    """Check that filter_path filter works as expected."""
    current_filter = filters.filter_ssl()
    http_request = f'GET {path} HTTP/1.1'
    line = line_factory(http_request=http_request)
    assert current_filter(line) is result


@pytest.mark.parametrize(('tr', 'result'), [(45, False), (13000, True), (4566, False)])
def test_filter_slow_requests(line_factory, tr, result):
    """Check that filter_slow_requests filter works as expected."""
    current_filter = filters.filter_slow_requests('10000')
    line = line_factory(tr=tr)
    assert current_filter(line) is result


@pytest.mark.parametrize(('tw', 'result'), [(45, False), (13000, True), (4566, True)])
def test_filter_wait_on_queues(line_factory, tw, result):
    """Check that filter_wait_on_queues filter works as expected"""
    current_filter = filters.filter_wait_on_queues('50')
    line = line_factory(tw=tw)
    assert current_filter(line) is result


@pytest.mark.parametrize(
    ('to_filter', 'to_check', 'result'),
    [
        ('200', '200', True),
        ('200', '230', False),
        ('300', '300', True),
        ('300', '400', False),
    ],
)
def test_filter_status_code(line_factory, to_filter, to_check, result):
    """Test that the status_code filter works as expected."""
    current_filter = filters.filter_status_code(to_filter)
    line = line_factory(status=to_check)
    assert current_filter(line) is result


@pytest.mark.parametrize(
    ('to_filter', 'to_check', 'result'),
    [
        ('2', '200', True),
        ('2', '230', True),
        ('2', '300', False),
        ('3', '300', True),
        ('3', '330', True),
        ('3', '400', False),
    ],
)
def test_filter_status_code_family(line_factory, to_filter, to_check, result):
    """Test that the status_code_family filter works as expected."""
    current_filter = filters.filter_status_code_family(to_filter)
    line = line_factory(status=to_check)
    assert current_filter(line) is result


@pytest.mark.parametrize(
    ('to_filter', 'to_check', 'result'),
    [
        ('GET', 'GET', True),
        ('GET', 'POST', False),
        ('GET', 'PUT', False),
        ('GET', 'PATCH', False),
        ('GET', 'DELETE', False),
        ('PATCH', 'PATCH', True),
        ('DELETE', 'DELETE', True),
    ],
)
def test_filter_http_method(line_factory, to_filter, to_check, result):
    """Test that the http_method filter works as expected."""
    current_filter = filters.filter_http_method(to_filter)
    line = line_factory(http_request=f'{to_check} /path HTTP/1.1')
    assert current_filter(line) is result


@pytest.mark.parametrize(
    ('to_filter', 'to_check', 'result'),
    [
        ('default', 'default', True),
        ('default', 'backend', False),
        ('backend', 'backend', True),
        ('backend', 'default', False),
    ],
)
def test_filter_backend(line_factory, to_filter, to_check, result):
    """Test that the backend filter works as expected."""
    current_filter = filters.filter_backend(to_filter)
    line = line_factory(backend_name=to_check)
    assert current_filter(line) is result


@pytest.mark.parametrize(
    ('to_filter', 'to_check', 'result'),
    [
        ('varnish', 'varnish', True),
        ('varnish', 'nginx', False),
        ('nginx', 'nginx', True),
        ('nginx', 'varnish', False),
    ],
)
def test_filter_frontend(line_factory, to_filter, to_check, result):
    """Test that the frontend filter works as expected."""
    current_filter = filters.filter_frontend(to_filter)
    line = line_factory(frontend_name=to_check)
    assert current_filter(line) is result


@pytest.mark.parametrize(
    ('to_filter', 'to_check', 'result'),
    [
        ('server1', 'server1', True),
        ('server1', 'backend23', False),
        ('backend23', 'backend23', True),
        ('backend23', 'server1', False),
    ],
)
def test_filter_server(line_factory, to_filter, to_check, result):
    """Test that the server filter works as expected."""
    current_filter = filters.filter_server(to_filter)
    line = line_factory(server_name=to_check)
    assert current_filter(line) is result


@pytest.mark.parametrize(
    ('to_filter', 'to_check', 'result'),
    [
        ('400', '500', True),
        ('400', '+500', True),
        ('+400', '500', True),
        ('+400', '+500', True),
        ('400', '300', False),
        ('400', '+300', False),
        ('+400', '300', False),
        ('+400', '+300', False),
    ],
)
def test_filter_response_size(line_factory, to_filter, to_check, result):
    """Test that the size filter works as expected.

    Note that both filter and value can have a leading plus sign.
    """
    current_filter = filters.filter_response_size(to_filter)
    line = line_factory(bytes=to_check)
    assert current_filter(line) is result



from datetime import datetime
from datetime import timedelta
from haproxy.utils import date_str_to_datetime
from haproxy.utils import delta_str_to_timedelta
from haproxy.utils import VALID_COMMANDS
from haproxy.utils import VALID_FILTERS
from haproxy.utils import validate_arg_date
from haproxy.utils import validate_arg_delta

import pytest


@pytest.mark.parametrize(
    ('text', 'expected'),
    [
        ('45s', timedelta(seconds=45)),
        ('2m', timedelta(minutes=2)),
        ('13h', timedelta(hours=13)),
        ('2d', timedelta(days=2)),
    ],
)
def test_str_to_timedelta(text, expected):
    """Check that deltas are converted to timedelta objects."""
    assert delta_str_to_timedelta(text) == expected


@pytest.mark.parametrize(
    ('text', 'expected'),
    [
        ('04/Jan/2013', datetime(2013, 1, 4)),
        ('13/May/2015:13', datetime(2015, 5, 13, 13)),
        ('22/Jun/2017:12:11', datetime(2017, 6, 22, 12, 11)),
        ('29/Aug/2019:10:09:08', datetime(2019, 8, 29, 10, 9, 8)),
    ],
)
def test_str_to_datetime(text, expected):
    """Check that start are converted to datetime objects."""
    assert date_str_to_datetime(text) == expected


@pytest.mark.parametrize('cmd_key', [*VALID_COMMANDS])
def test_valid_commands(cmd_key):
    """Check that the commands' information is complete."""
    cmd_data = VALID_COMMANDS[cmd_key]
    assert cmd_data['klass']
    assert cmd_data['klass'].command_line_name() == cmd_key
    assert cmd_data['description']
    assert cmd_data['description'].startswith(f'{cmd_key}:\n\t')


@pytest.mark.parametrize('filter_key', [*VALID_FILTERS])
def test_valid_filters(filter_key):
    """Check that the filters' information is complete."""
    filter_data = VALID_FILTERS[filter_key]
    assert filter_data['obj']
    assert filter_data['obj'].__name__ == f'filter_{filter_key}'
    assert filter_data['description']
    assert filter_data['description'].startswith(f'{filter_key}:\n\t')


@pytest.mark.parametrize(('value', 'expected'), [('', None), ('30/Dec/2019', True)])
def test_validate_date(value, expected):
    """Check that the date is validated or an exception raised."""
    if expected is None:
        with pytest.raises(ValueError, match='--start argument is not valid'):
            validate_arg_date(value)
    else:
        assert validate_arg_date(value) is None


@pytest.mark.parametrize(('value', 'expected'), [('', None), ('3d', True)])
def test_validate_delta(value, expected):
    """Check that the delta is validated or an exception raised."""
    if expected is None:
        with pytest.raises(ValueError, match='--delta argument is not valid'):
            validate_arg_delta(value)

    else:
        assert validate_arg_delta(value) is None



from datetime import datetime
from datetime import timedelta

import pytest


NOW = datetime.now()
TWO_DAYS_AGO = NOW - timedelta(days=2)
IN_TWO_DAYS = NOW + timedelta(days=2)


def test_default_values(line_factory, default_line_data):
    line = line_factory()

    assert line.client_ip == default_line_data['client_ip']
    assert line.client_port == default_line_data['client_port']

    assert line.raw_accept_date in default_line_data['accept_date']

    assert line.frontend_name == default_line_data['frontend_name']
    assert line.backend_name == default_line_data['backend_name']
    assert line.server_name == default_line_data['server_name']

    assert line.time_wait_request == default_line_data['tq']
    assert line.time_wait_queues == default_line_data['tw']
    assert line.time_connect_server == default_line_data['tc']
    assert line.time_wait_response == default_line_data['tr']
    assert line.total_time == default_line_data['tt']

    assert line.status_code == default_line_data['status']
    assert line.bytes_read == default_line_data['bytes']

    assert line.connections_active == default_line_data['act']
    assert line.connections_frontend == default_line_data['fe']
    assert line.connections_backend == default_line_data['be']
    assert line.connections_server == default_line_data['srv']
    assert line.retries == default_line_data['retries']

    assert line.queue_server == default_line_data['queue_server']
    assert line.queue_backend == default_line_data['queue_backend']

    assert line.captured_request_headers == default_line_data['headers'].strip()[1:-1]
    assert line.captured_response_headers is None

    assert line.raw_http_request == default_line_data['http_request']

    assert line.is_valid


def test_unused_values(line_factory):
    line = line_factory()
    assert line.captured_request_cookie is None
    assert line.captured_response_cookie is None
    assert line.termination_state is None


def test_datetime_value(line_factory):
    line = line_factory()
    assert isinstance(line.accept_date, datetime)


def test_http_request_values(line_factory):
    method = 'PUT'
    path = '/path/to/my/image'
    protocol = 'HTTP/2.0'
    line = line_factory(http_request=f'{method} {path} {protocol}')
    assert line.http_request_method == method
    assert line.http_request_path == path
    assert line.http_request_protocol == protocol


def test_invalid_line(line_factory):
    line = line_factory(bytes='wroooong')
    assert not line.is_valid


def test_no_captured_headers(line_factory):
    """A log line without captured headers is still valid."""
    line = line_factory(headers='')
    assert line.is_valid


def test_request_and_response_captured_headers(line_factory):
    """Request and response headers captured are parsed correctly."""
    request_headers = '{something}'
    response_headers = '{something_else}'
    line = line_factory(headers=f' {request_headers} {response_headers}')
    assert line.is_valid
    assert f'{{{line.captured_request_headers}}}' == request_headers
    assert f'{{{line.captured_response_headers}}}' == response_headers


def test_request_is_https_valid(line_factory):
    """Check that if a log line contains the SSL port on it, is reported
    as a https connection.
    """
    line = line_factory(http_request='GET /domain:443/to/image HTTP/1.1')
    assert line.is_https


def test_request_is_https_false(line_factory):
    """Check that if a log line does not contains the SSL port on it, is
    not reported as a https connection.
    """
    line = line_factory(http_request='GET /domain:80/to/image HTTP/1.1')
    assert not line.is_https


def test_request_is_front_page(line_factory):
    """Check that if a request is for the front page the request path is
    correctly stored.
    """
    line = line_factory(http_request='GET / HTTP/1.1')
    assert line.http_request_path == '/'


@pytest.mark.parametrize(
    'process',
    [
        'ip-192-168-1-1 haproxy[28029]:',
        'dvd-ctrl1 haproxy[403100]:',
        'localhost.localdomain haproxy[2345]:',
    ],
)
def test_process_names(line_factory, process):
    """Checks that different styles of process names are handled correctly."""
    line = line_factory(process_name_and_pid=process)
    assert line.is_valid is True


def test_unparseable_http_request(line_factory):
    line = line_factory(http_request='something')
    assert line.http_request_method == 'invalid'
    assert line.http_request_path == 'invalid'
    assert line.http_request_protocol == 'invalid'


def test_truncated_requests(line_factory):
    """Check that truncated requests are still valid.

    That would be requests that do not have the protocol part specified.
    """
    line = line_factory(http_request='GET /')
    assert line.http_request_method == 'GET'
    assert line.http_request_path == '/'
    assert line.http_request_protocol is None


@pytest.mark.parametrize(
    'syslog',
    [
        # nixos format
        '2017-07-06T14:29:39+02:00',
        # regular format
        'Dec  9 13:01:26',
    ],
)
def test_syslog(line_factory, syslog):
    """Check that the timestamp at the beginning are parsed.

    We support different syslog formats, NixOS style and the one on other Linux.
    """
    line = line_factory(syslog_date=syslog)
    assert line.is_valid is True


def test_ip_from_headers(line_factory):
    """Check that the IP from the captured headers takes precedence."""
    line = line_factory(headers=' {1.2.3.4}')
    assert line.ip == '1.2.3.4'


@pytest.mark.parametrize(
    'ip',
    ['1.2.3.4', '1.2.3.4, 2.3.4.5', '1.2.3.4,2.3.4.5,5.4.3.2'],
)
def test_only_first_ip_from_headers(line_factory, ip):
    """Check that if there are multiple IPs, only the first one is used."""
    line = line_factory(headers=f' {{{ip}}}')
    assert line.ip == '1.2.3.4'


@pytest.mark.parametrize(
    'ip',
    ['127.1.2.7', '1.127.230.47', 'fe80::9379:c29e:6701:cef8', 'fe80::9379:c29e::'],
)
def test_ip_from_client_ip(line_factory, ip):
    """Check that if there is no IP on the captured headers, the client IP is used."""
    line = line_factory(headers='', client_ip=ip)
    assert line.ip == ip


@pytest.mark.parametrize(
    ('start', 'end', 'result'),
    [
        (None, None, True),
        (TWO_DAYS_AGO, None, True),
        (IN_TWO_DAYS, None, False),
        (TWO_DAYS_AGO, IN_TWO_DAYS, True),
        (TWO_DAYS_AGO, TWO_DAYS_AGO, False),
    ],
)
def test_is_within_timeframe(line_factory, start, end, result):
    """Check that a line is within a given time frame."""
    line = line_factory(accept_date=NOW.strftime('%d/%b/%Y:%H:%M:%S.%f'))
    assert line.is_within_time_frame(start, end) is result



from haproxy.logfile import Log
from haproxy.utils import VALID_COMMANDS
from haproxy.utils import VALID_FILTERS
from haproxy.utils import validate_arg_date
from haproxy.utils import validate_arg_delta

import argparse
import os


def create_parser():
    desc = 'Analyze HAProxy log files and outputs statistics about it'
    parser = argparse.ArgumentParser(description=desc)

    parser.add_argument('-l', '--log', help='HAProxy log file to analyze')

    parser.add_argument(
        '-s',
        '--start',
        help='Process log entries starting at this time, in HAProxy date '
        'format (e.g. 11/Dec/2013 or 11/Dec/2013:19:31:41). '
        'At least provide the day/month/year. Values not specified will '
        'use their base value (e.g. 00 for hour). Use in conjunction '
        'with -d to limit the number of entries to process.',
    )

    parser.add_argument(
        '-d',
        '--delta',
        help='Limit the number of entries to process. Express the time delta '
        'as a number and a time unit, e.g.: 1s, 10m, 3h or 4d (for 1 '
        'second, 10 minutes, 3 hours or 4 days). Use in conjunction with '
        '-s to only analyze certain time delta. If no start time is '
        'given, the time on the first line will be used instead.',
    )

    parser.add_argument(
        '-c',
        '--command',
        help='List of commands, comma separated, to run on the log file. See '
        '--list-commands to get a full list of them.',
    )

    parser.add_argument(
        '-f',
        '--filter',
        help='List of filters to apply on the log file. Passed as comma '
        'separated and parameters within square brackets, e.g '
        'ip[192.168.1.1],ssl,path[/some/path]. See '
        '--list-filters to get a full list of them.',
    )

    parser.add_argument(
        '-n',
        '--negate-filter',
        help='Make filters passed with -f work the other way around, i.e. if '
        'the ``ssl`` filter is passed instead of showing only ssl '
        'requests it will show non-ssl traffic. If the ``ip`` filter is '
        'used, then all but that ip passed to the filter will be used.',
        action='store_true',
    )

    parser.add_argument(
        '--list-commands', action='store_true', help='Lists all commands available.'
    )

    parser.add_argument(
        '--list-filters', action='store_true', help='Lists all filters available.'
    )

    parser.add_argument('--json', action='store_true', help='Output results in json.')
    parser.add_argument(
        '--invalid',
        action='store_false',
        help='Print the lines that could not be parsed. '
        'Be aware that mixing it with the print command will mix their output.',
    )

    return parser


def parse_arguments(args):
    data = {
        'start': None,
        'delta': None,
        'commands': None,
        'filters': None,
        'negate_filter': None,
        'log': None,
        'list_commands': None,
        'list_filters': None,
        'json': None,
        'invalid_lines': None,
    }

    if args.list_commands:
        data['list_commands'] = True
        # no need to further process any other input parameter
        return data

    if args.list_filters:
        data['list_filters'] = True
        # no need to further process any other input parameter
        return data

    if args.negate_filter:
        data['negate_filter'] = True

    if args.start is not None:
        validate_arg_date(args.start)
        data['start'] = args.start

    if args.delta is not None:
        validate_arg_delta(args.delta)
        data['delta'] = args.delta

    if args.command is not None:
        data['commands'] = parse_arg_commands(args.command)

    if args.filter is not None:
        data['filters'] = parse_arg_filters(args.filter)

    if args.log is not None:
        _validate_arg_logfile(args.log)
        data['log'] = args.log

    if args.json is not None:
        data['json'] = args.json

    if args.invalid:
        data['invalid_lines'] = args.json

    return data


def parse_arg_commands(commands_list):
    input_commands = commands_list.split(',')
    for cmd in input_commands:
        if cmd not in VALID_COMMANDS:
            raise ValueError(
                f'command "{cmd}" is not available. '
                'Use --list-commands to get a list of all available commands.'
            )
    return input_commands


def parse_arg_filters(filters_arg):
    input_filters = filters_arg.split(',')

    return_data = []
    for filter_expression in input_filters:
        filter_name = filter_expression
        filter_arg = None

        if filter_expression.endswith(']'):
            if '[' not in filter_expression:
                raise ValueError(
                    f'Error on filter "{filter_expression}". '
                    f'It is missing an opening square bracket.'
                )
            filter_name, filter_arg = filter_expression.split('[')
            filter_arg = filter_arg[:-1]  # remove the closing square bracket

        if filter_name not in VALID_FILTERS:
            raise ValueError(
                f'filter "{filter_name}" is not available. Use --list-filters to get a list of all available filters.'
            )

        return_data.append((filter_name, filter_arg))

    return return_data


def _validate_arg_logfile(filename):
    filepath = os.path.join(os.getcwd(), filename)
    if not os.path.exists(filepath):
        raise ValueError(f'filename {filepath} does not exist')


def print_commands():
    """Prints all commands available with their description."""
    for command_name in sorted(VALID_COMMANDS.keys()):
        print(VALID_COMMANDS[command_name]['description'])


def print_filters():
    """Prints all filters available with their description."""
    for filter_name in sorted(VALID_FILTERS.keys()):
        print(VALID_FILTERS[filter_name]['description'])


def show_help(data):
    # make sure that if no arguments are passed the help is shown
    show = True
    ignore_keys = ('log', 'json', 'negate_filter', 'invalid_lines')
    for key in data:
        if data[key] is not None and key not in ignore_keys:
            show = False
            break

    if show:
        parser = create_parser()
        parser.print_help()
        return True
    return False


def main(args):
    if show_help(args):
        return

    # show the command list
    if args['list_commands']:
        print_commands()
        # no need to process further
        return

    # show the filter list
    if args['list_filters']:
        print_filters()
        # no need to process further
        return

    # initialize the log file
    log_file = Log(
        logfile=args['log'],
        start=args['start'],
        delta=args['delta'],
        show_invalid=args['invalid_lines'],
    )

    # get the commands and filters to use
    filters_to_use = requested_filters(args)
    cmds_to_use = requested_commands(args)

    # double negation: when a user wants to negate the filters,
    # the argument parsing sets `negate_filter` to True,
    # but the filtering logic (the `all()`) returns True if the line meets all filters
    # so reversing whatever `negate_filter` has is what the user wants :)
    expected_filtering = True
    if args['negate_filter']:
        expected_filtering = False
    # process all log lines
    for line in log_file:
        if all(f(line) for f in filters_to_use) is expected_filtering:
            for cmd in cmds_to_use:
                cmd(line)

    # print the results
    print('\nRESULTS\n')
    output = None
    if args['json']:
        output = 'json'
    for cmd in cmds_to_use:
        cmd.results(output=output)


def requested_filters(args):
    filters_list = []
    if args['filters']:
        for filter_name, arg in args['filters']:
            filter_func = VALID_FILTERS[filter_name]['obj']
            filters_list.append(filter_func(arg))
    return filters_list


def requested_commands(args):
    cmds_list = []
    for command in args['commands']:
        cmd_klass = VALID_COMMANDS[command]['klass']
        cmds_list.append(cmd_klass())
    return cmds_list


def console_script():  # pragma: no cover
    parser = create_parser()
    arguments = parse_arguments(parser.parse_args())
    main(arguments)



from haproxy.main import create_parser
from haproxy.main import main
from haproxy.main import parse_arguments
from haproxy.utils import VALID_COMMANDS
from haproxy.utils import VALID_FILTERS

import pytest
import sys


PY310_OR_HIGHER = sys.version_info[1] > 9


@pytest.fixture()
def default_arguments():
    """Return all the expected arguments the main function expects."""
    return {
        'start': None,
        'delta': None,
        'log': 'tests/files/small.log',
        'commands': ['counter'],
        'negate_filter': None,
        'filters': None,
        'list_commands': False,
        'list_filters': False,
        'json': False,
        'invalid_lines': False,
    }


@pytest.mark.parametrize(
    ('switch', 'listing'),
    [('list-filters', VALID_FILTERS), ('list-commands', VALID_COMMANDS)],
)
def test_list_filters_and_commands(capsys, switch, listing):
    """Test that one can request the filters/commands to be listed."""
    parser = create_parser()
    data = parse_arguments(parser.parse_args([f'--{switch}']))
    argument = switch.replace('-', '_')
    for key in data:
        expected = None
        if key == argument:
            expected = True
        assert data[key] is expected
    main(data)
    output_text = capsys.readouterr().out
    for name in listing:
        assert f'{name}:\n\t' in output_text


def test_show_help(capsys):
    """Check that the help is shown if no arguments are given."""
    parser = create_parser()
    data = parse_arguments(parser.parse_args([]))
    main(data)
    output_text = capsys.readouterr().out
    if PY310_OR_HIGHER:
        assert 'options:' in output_text
    else:
        assert 'optional arguments:' in output_text
    assert '--list-filters ' in output_text
    assert '--list-commands ' in output_text


def test_main(capsys, default_arguments):
    """Check that the main function works as expected with default arguments."""
    main(default_arguments)
    output_text = capsys.readouterr().out
    assert 'COUNTER\n=======\n9' in output_text


def test_main_with_filter(capsys, default_arguments):
    """Check that the filters are applied as expected."""
    default_arguments['filters'] = [
        ('server', 'instance1'),
    ]
    main(default_arguments)
    output_text = capsys.readouterr().out
    assert 'COUNTER\n=======\n4' in output_text


def test_main_negate_filter(capsys, default_arguments):
    """Check that filters can be reversed."""
    default_arguments['filters'] = [
        ('server', 'instance1'),
    ]
    default_arguments['negate_filter'] = True
    main(default_arguments)
    output_text = capsys.readouterr().out
    assert 'COUNTER\n=======\n5' in output_text


def test_print_no_output(capsys, default_arguments):
    """Check that the print header is not shown."""
    default_arguments['commands'] = ['print']
    main(default_arguments)
    output_text = capsys.readouterr().out
    assert 'PRINT\n=====' not in output_text


def test_json_output(capsys, default_arguments):
    """Check that the JSON switch is used and JSON output is printed."""
    default_arguments['json'] = True
    main(default_arguments)
    output_text = capsys.readouterr().out
    assert 'COUNTER\n=======\n9' not in output_text
    assert '{"COUNTER": 9}' in output_text



from datetime import datetime
from haproxy.logfile import Log

import pytest


def test_logfile_default_values():
    """Check that the default values are set."""
    log_file = Log('something')
    assert log_file.logfile == 'something'
    assert log_file.show_invalid is False
    assert log_file.invalid_lines == 0
    assert log_file.valid_lines == 0
    assert log_file.total_lines == 0
    assert log_file.start is None
    assert log_file.end is None


@pytest.mark.parametrize(
    ('start_str', 'start_obj', 'delta', 'end_obj'),
    [
        (None, None, None, None),
        (None, None, '3d', None),
        ('12/Dec/2019', datetime(2019, 12, 12), None, None),
        ('12/Dec/2019', datetime(2019, 12, 12), '3d', datetime(2019, 12, 15)),
    ],
)
def test_start_and_end_attributes(start_str, start_obj, delta, end_obj):
    """Check that the start and end of attributes of Log objects are set as expected."""
    log_file = Log('something', start=start_str, delta=delta)
    assert log_file.logfile == 'something'
    assert log_file.invalid_lines == 0
    assert log_file.start == start_obj
    assert log_file.end == end_obj


@pytest.mark.parametrize('accept_date', ['09/Dec/2013:12:59:46.633', None])
def test_lines_validity(tmp_path, line_factory, accept_date):
    """Check that lines are either counted as valid or invalid."""
    file_path = tmp_path / 'haproxy.log'
    line = ''
    if accept_date:
        line = line_factory(accept_date=accept_date).raw_line
    with open(file_path, 'w') as file_obj:
        file_obj.write(f'{line}\n')
    log_file = Log(file_path)
    _ = list(log_file)

    assert log_file.total_lines == 1
    if accept_date:
        assert log_file.valid_lines == 1
        assert log_file.invalid_lines == 0
    else:
        assert log_file.valid_lines == 0
        assert log_file.invalid_lines == 1


@pytest.mark.parametrize(
    ('accept_date', 'start', 'delta', 'is_valid'),
    [
        # valid line and no time frame, returned
        ('09/Dec/2013:12:59:46.633', None, None, True),
        # invalid line, not returned
        (None, None, None, False),
        # valid line before time frame, not returned
        ('09/Dec/2013:12:59:46.633', '09/Dec/2014', None, False),
        # valid line after time frame, not returned
        ('09/Dec/2013:12:59:46.633', '08/Dec/2012', '3d', False),
        # valid line within time frame, returned
        ('09/Dec/2013:12:59:46.633', '08/Dec/2013', '3d', True),
    ],
)
def test_returned_lines(tmp_path, line_factory, accept_date, start, delta, is_valid):
    """Check that lines are only returned if they are valid AND within the time frame."""
    file_path = tmp_path / 'haproxy.log'
    line = ''
    if accept_date:
        line = line_factory(accept_date=accept_date).raw_line
    with open(file_path, 'w') as file_obj:
        file_obj.write(f'{line}\n')
    log_file = Log(file_path, start=start, delta=delta)
    lines = list(log_file)
    assert bool(len(lines)) is is_valid


def test_total_lines():
    """Check that the total amount of lines are always counted."""
    log_file = Log(logfile='tests/files/2_ok_1_invalid.log')
    _ = list(log_file)
    assert log_file.total_lines == 3
    assert log_file.valid_lines == 2
    assert log_file.invalid_lines == 1


@pytest.mark.parametrize('headers', [' {1.2.3.4}', 'random-value-that-breaks'])
def test_print_invalid_lines(tmp_path, line_factory, headers, capsys):
    """Check that invalid lines are printed, if asked to do so."""
    file_path = tmp_path / 'haproxy.log'
    line = line_factory(headers=headers).raw_line
    with open(file_path, 'w') as file_obj:
        file_obj.write(f'{line}\n')
    log_file = Log(file_path, show_invalid=True)
    _ = list(log_file)

    output = capsys.readouterr().out
    if log_file.valid_lines == 1:
        assert headers not in output
    else:
        assert headers in output



from haproxy.main import create_parser
from haproxy.main import parse_arg_filters
from haproxy.main import parse_arguments

import pytest


def test_parser_arguments_defaults():
    """Test that the argument parsing defaults works."""
    parser = create_parser()
    data = parse_arguments(parser.parse_args([]))
    assert data == {
        'start': None,
        'delta': None,
        'commands': None,
        'filters': None,
        'negate_filter': None,
        'log': None,
        'list_commands': None,
        'list_filters': None,
        'json': False,
        'invalid_lines': False,
    }


@pytest.mark.parametrize(
    ('argument', 'option'),
    [
        ('--list-commands', 'list_commands'),
        ('--list-filters', 'list_filters'),
        ('--negate-filter', 'negate_filter'),
        ('-n', 'negate_filter'),
        ('--json', 'json'),
    ],
)
def test_parser_boolean_arguments(argument, option):
    """Test that the argument parsing defaults works."""
    parser = create_parser()
    data = parse_arguments(parser.parse_args([argument]))
    assert data[option] is True


@pytest.mark.parametrize(
    ('start', 'delta'), [('30/Dec/2019', '3d'), ('20/Jun/2015', '2h')]
)
def test_arguments_dates(start, delta):
    """Check that properly formatted start and delta arguments are processed fine.

    Thus they are extracted and stored for later use.
    """
    parser = create_parser()
    data = parse_arguments(parser.parse_args(['-s', start, '-d', delta]))
    assert data['start'] == start
    assert data['delta'] == delta


@pytest.mark.parametrize('start', ['33/Dec/2019', '5/Hallo/2019'])
def test_arguments_date_invalid(start):
    """Incorrectly formatted start argument raises an exception."""
    parser = create_parser()
    with pytest.raises(ValueError, match='--start argument is not valid'):
        parse_arguments(parser.parse_args(['-s', start]))


@pytest.mark.parametrize('delta', ['3P', '2323MM'])
def test_arguments_delta_invalid(delta):
    """Incorrectly formatted delta argument raises an exception."""
    parser = create_parser()
    with pytest.raises(ValueError, match='--delta argument is not valid'):
        parse_arguments(parser.parse_args(['-d', delta]))


@pytest.mark.parametrize(
    ('cmds', 'is_valid'),
    [
        ('counter', True),
        ('counter,ip_counter', True),
        ('ip_counter,count_data', False),
        ('count_data', False),
    ],
)
def test_commands_arguments(cmds, is_valid):
    """Test that the commands are parsed, and an exception raised otherwise."""
    parser = create_parser()
    if not is_valid:
        with pytest.raises(ValueError, match='is not available. Use --list-commands'):
            parse_arguments(parser.parse_args(['-c', cmds]))
    else:
        data = parse_arguments(parser.parse_args(['-c', cmds]))
        assert data['commands'] == cmds.split(',')


@pytest.mark.parametrize(
    ('filters_list', 'is_valid'),
    [
        ('ip_range', True),
        ('slow_requests,backend', True),
        ('tomatoes', False),
        ('slow_requests,potatoes', False),
    ],
)
def test_filters_arguments(filters_list, is_valid):
    """Test that the filters are parsed, and an exception raised otherwise."""
    parser = create_parser()
    if not is_valid:
        with pytest.raises(ValueError, match='is not available. Use --list-filters'):
            parse_arguments(parser.parse_args(['-f', filters_list]))
    else:
        data = parse_arguments(parser.parse_args(['-f', filters_list]))
        assert data['filters'] == [(x, None) for x in filters_list.split(',')]


@pytest.mark.parametrize(
    ('filter_expression', 'expected'),
    [
        ('ip_range', [('ip_range', None)]),
        ('ip_rangelala]', None),
        ('ip_range[lala]', [('ip_range', 'lala')]),
    ],
)
def test_filters_with_arguments(filter_expression, expected):
    """Check that the arguments given to the filters are parsed properly.

    Or raise and exception otherwise.
    """
    if expected is None:
        with pytest.raises(ValueError, match='It is missing an opening square bracket'):
            parse_arg_filters(filter_expression)
    else:
        data = parse_arg_filters(filter_expression)
        assert data == expected


@pytest.mark.parametrize(
    ('filename', 'is_valid'),
    [
        ('tests/conftest.py', True),
        ('tests/non-existing-file.py', False),
    ],
)
def test_log_argument(filename, is_valid):
    """Check that the argument parsing validates that the file exists."""
    parser = create_parser()
    if is_valid:
        data = parse_arguments(parser.parse_args(['-l', filename]))
        assert data['log'] == filename
    else:
        with pytest.raises(ValueError, match=f'{filename} does not exist'):
            parse_arguments(parser.parse_args(['-l', filename]))



from collections import defaultdict
from collections import OrderedDict
from datetime import datetime

import json
import time


class BaseCommandMixin:
    @classmethod
    def command_line_name(cls):
        """Convert class name to lowercase with underscores.

        i.e. turn HttpMethods to http_methods.
        """
        final_string = cls.__name__[0].lower()
        for character in cls.__name__[1:]:
            if character.isupper():
                final_string += f'_{character.lower()}'
            else:
                final_string += character
        return final_string

    def raw_results(self):  # pragma: no cover
        raise NotImplementedError

    def json_data(self):
        return self.raw_results()

    def print_data(self):
        return self.raw_results()

    def results(self, output=None):
        command_name = self.command_line_name().upper()
        if output == 'json':
            results = self.json_data()
            print(json.dumps({command_name: results}))
        else:
            results = self.print_data()
            underline = '=' * len(command_name)
            print(f'{command_name}\n{underline}\n{results}\n')


class AttributeCounterMixin:
    attribute_name = None

    def __init__(self):
        self.stats = defaultdict(int)

    def __call__(self, line):
        self.stats[getattr(line, self.attribute_name)] += 1

    def raw_results(self):
        return self.stats

    def print_data(self):
        result = ''
        data = self.raw_results()
        if isinstance(data, list):
            data = dict(data)
        data = sorted(data.items(), key=lambda data_info: data_info[1], reverse=True)
        for key, value in data:
            result += f'- {key}: {value}\n'
        return result

    def json_data(self):
        result = []
        data = sorted(
            self.stats.items(), key=lambda data_info: data_info[1], reverse=True
        )
        for key, value in data:
            result.append({key: value})
        return result


class SortTrimMixin:
    @staticmethod
    def _sort_and_trim(data, reverse=False):
        """Sorts a dictionary with at least two fields on each of them sorting
        by the second element.

        .. warning::
          Right now is hardcoded to 10 elements, improve the command line
          interface to allow to send parameters to each command or globally.
        """
        threshold = 10
        data_list = data.items()
        data_list = sorted(
            data_list, key=lambda data_info: data_info[1], reverse=reverse
        )
        return data_list[:threshold]


class Counter(BaseCommandMixin):
    """Count valid lines."""

    def __init__(self):
        self.counter = 0

    def __call__(self, line):
        self.counter += 1

    def raw_results(self):
        return self.counter


class HttpMethods(AttributeCounterMixin, BaseCommandMixin):
    """Tally all requests per HTTP method (GET/POST...)."""

    attribute_name = 'http_request_method'


class IpCounter(AttributeCounterMixin, BaseCommandMixin):
    """Report a breakdown of how many requests have been made per IP.

    For this to work you need to configure HAProxy to capture
    the `X-Forwarded-For` header.
    """

    attribute_name = 'ip'


class TopIps(IpCounter, SortTrimMixin):
    """Return the top most frequent IPs (10 items)."""

    def raw_results(self):
        return self._sort_and_trim(self.stats, reverse=True)


class StatusCodesCounter(AttributeCounterMixin, BaseCommandMixin):
    """Tally requests per HTTP status (404, 500...)"""

    attribute_name = 'status_code'


class RequestPathCounter(AttributeCounterMixin, BaseCommandMixin):
    """Tally requests per the request's path."""

    attribute_name = 'http_request_path'


class TopRequestPaths(RequestPathCounter, SortTrimMixin):
    """Returns the top most frequent paths (10 items)."""

    def raw_results(self):
        return self._sort_and_trim(self.stats, reverse=True)


class SlowRequests(BaseCommandMixin):
    """List all requests that are considered slow to process (1 second)."""

    threshold = 1000

    def __init__(self):
        self.slow_requests = []

    def __call__(self, line):
        response_time = line.time_wait_response
        if response_time >= self.threshold:
            self.slow_requests.append(response_time)

    def raw_results(self):
        return sorted(self.slow_requests)


class SlowRequestsCounter(SlowRequests):
    """Counts requests that are considered slow (1 second)."""

    def raw_results(self):
        return len(self.slow_requests)


class AverageResponseTime(SlowRequests):
    """Global average response time it took downstream servers to answer requests."""

    threshold = 0

    def raw_results(self):
        total_requests = float(len(self.slow_requests))
        if total_requests > 0:
            average = sum(self.slow_requests) / total_requests
            return round(average, 2)
        return 0.0


class AverageWaitingTime(BaseCommandMixin):
    """Return the average time valid requests wait on HAProxy before being dispatched to a backend server."""

    def __init__(self):
        self.waiting_times = []

    def __call__(self, line):
        waiting_time = line.time_wait_queues
        if waiting_time >= 0:
            self.waiting_times.append(waiting_time)

    def raw_results(self):
        total_requests = float(len(self.waiting_times))
        if total_requests > 0:
            average = sum(self.waiting_times) / total_requests
            return round(average, 2)
        return 0.0


class ServerLoad(AttributeCounterMixin, BaseCommandMixin):
    """Tally requests per downstream server."""

    attribute_name = 'server_name'


class QueuePeaks(BaseCommandMixin):
    """Give stats about queue peaks in HAProxy.

    When servers can not handle all incoming requests, they have to wait on HAProxy.
    On every log line there is an account for how many requests have been piled up.

    A queue peak is defined by the biggest value on the backend queue
    on a series of log lines that are between log lines with the queue empty.
    """

    def __init__(self):
        self.requests = {}
        self.threshold = 1

    @staticmethod
    def _generate_key(date):
        """Create a suitable unique key out of a python datetime.datetime object."""
        # get the unix timestamp out of the date,
        # after removing the microseconds from it
        no_microseconds = date.replace(microsecond=0)
        time_parts = no_microseconds.timetuple()
        unixtime = time.mktime(time_parts)

        # add back the microseconds to the key, as decimals
        microseconds = date.microsecond / (10 ** len(str(date.microsecond)))
        key = unixtime + microseconds
        return key

    def __call__(self, line):
        key = self._generate_key(line.accept_date)
        self.requests[key] = (line.queue_backend, line.accept_date)

    def raw_results(self):
        sorted_requests = OrderedDict(sorted(self.requests.items()))
        peaks = []
        current_peak = 0
        requests_on_queue = 0
        timestamp = None

        current_span = 0
        first_with_queue = None

        for requests_on_queue, timestamp in sorted_requests.values():
            # set the peak
            if requests_on_queue > current_peak:
                current_peak = requests_on_queue

            # set the span
            if requests_on_queue > 0:
                current_span += 1

                # set when the queue starts
                if first_with_queue is None:
                    first_with_queue = timestamp

            # if the queue is already flushed, record it and reset values
            if requests_on_queue == 0 and current_peak > self.threshold:
                data = {
                    'peak': current_peak,
                    'span': current_span,
                    'started': first_with_queue,
                    'finished': timestamp,
                }
                peaks.append(data)
                current_peak = 0
                current_span = 0
                first_with_queue = None

        # case of a series that does not end
        if requests_on_queue > 0 and current_peak > self.threshold:
            data = {
                'peak': current_peak,
                'span': current_span,
                'started': first_with_queue,
                'finished': timestamp,
            }
            peaks.append(data)

        return peaks

    def print_data(self):
        data = ''
        for peak_info in self.raw_results():
            data += f'- peak: {peak_info.get("peak")} '  # noqa: Q000
            data += f'- span: {peak_info.get("span")} '  # noqa: Q000
            data += f'- started: {peak_info.get("started").isoformat()} '  # noqa: Q000
            data += (
                f'- finished: {peak_info.get("finished").isoformat()}\n'  # noqa: Q000
            )
        return data

    def json_data(self):
        data = self.raw_results()
        for peak_info in data:
            peak_info['started'] = peak_info['started'].isoformat()
            peak_info['finished'] = peak_info['finished'].isoformat()
        return data


class ConnectionType(BaseCommandMixin):
    """Tally requests per their SSL usage (either yes or no).

    This only works if the request path contains the default port for SSL (443).
    """

    def __init__(self):
        self.https = 0
        self.non_https = 0

    def __call__(self, line):
        if line.is_https:
            self.https += 1
        else:
            self.non_https += 1

    def raw_results(self):
        return self.https, self.non_https

    def print_data(self):
        https, http = self.raw_results()
        return f'- https: {https}\n- http: {http}'

    def json_data(self):
        https, http = self.raw_results()
        return [{'https': https}, {'http': http}]


class RequestsPerMinute(BaseCommandMixin):
    """Report the count of requests per minute.

    Combine it with time constrains (`-s` and `-d`) otherwise the output will be long.
    """

    def __init__(self):
        self.requests = defaultdict(int)

    def generate_key(self, accept_date):
        date_with_minute_precision = accept_date.replace(second=0, microsecond=0)
        unixtime = time.mktime(date_with_minute_precision.timetuple())
        return unixtime

    def __call__(self, line):
        key = self.generate_key(line.accept_date)
        self.requests[key] += 1

    def raw_results(self):
        """Return the list of requests sorted by the timestamp."""
        data = sorted(self.requests.items(), key=lambda data_info: data_info[0])
        return data

    def print_data(self):
        data = ''
        for date_info, count in self.raw_results():
            date = datetime.fromtimestamp(date_info).isoformat()
            data += f'- {date}: {count}\n'
        return data

    def json_data(self):
        data = []
        for date_info, count in self.raw_results():
            date = datetime.fromtimestamp(date_info).isoformat()
            data.append({date: count})
        return data


class RequestsPerHour(RequestsPerMinute):
    """Report the count of requests per hour.

    Combine it with time constrains (`-s` and `-d`) otherwise the output will be long.
    """

    def generate_key(self, accept_date):
        date_with_hour_precision = accept_date.replace(
            minute=0, second=0, microsecond=0
        )
        unixtime = time.mktime(date_with_hour_precision.timetuple())
        return unixtime


class Print(BaseCommandMixin):
    """Returns the raw lines to be printed."""

    def __call__(self, line):
        print(line.raw_line)

    def raw_results(self):
        return

    def results(self, output=None):
        return



def filter_ip(ip):
    """Filter by IP.

    -f ip[192.168.1.2]  # will return only lines that have this IP.

    Either the client IP, or, if present, the first IP captured
    in the X-Forwarded-For header.
    """

    def filter_func(log_line):
        return log_line.ip == ip

    return filter_func


def filter_ip_range(ip_range):
    """Filter by an IP range.

    -f ip_range[192.168.1]

    Rather than proper IP ranges, is a string matching.
    See `ip` filter about which IP is being.
    """

    def filter_func(log_line):
        ip = log_line.ip
        if ip:
            return ip.startswith(ip_range)

    return filter_func


def filter_path(path):
    """Filter by the request path.

    -f path[/one/two]

    It looks for the given path to be part of the requested path.
    """

    def filter_func(log_line):
        return path in log_line.http_request_path

    return filter_func


def filter_ssl(ignore=True):
    """Filter by SSL connection.

    -f ssl

    It checks that the request is made via the standard https port.
    """

    def filter_func(log_line):
        return log_line.is_https

    return filter_func


def filter_slow_requests(slowness):
    """Filter by response time.

    -f slow_requests[1000]  # get all lines that took more than a second to process

    Filters by the time it took the downstream server to process the request.
    Time is in milliseconds.
    """

    def filter_func(log_line):
        slowness_int = int(slowness)
        return slowness_int <= log_line.time_wait_response

    return filter_func


def filter_wait_on_queues(max_waiting):
    """Filter by queue time in HAProxy.

    -f wait_on_queues[1000]  # get all requests that waited more than a second in HAProxy

    Filters by the time a request had to wait in HAProxy
    prior to be sent to a downstream server to be processed.
    """

    def filter_func(log_line):
        waiting = int(max_waiting)
        return waiting <= log_line.time_wait_queues

    return filter_func


def filter_status_code(http_status):
    """Filter by a specific HTTP status code.

    -f status_code[404]
    """

    def filter_func(log_line):
        return log_line.status_code == http_status

    return filter_func


def filter_status_code_family(family_number):
    """Filter by a family of HTTP status code.

    -f status_code_family[5]  # get all 5xx status codes
    """

    def filter_func(log_line):
        return log_line.status_code.startswith(family_number)

    return filter_func


def filter_http_method(http_method):
    """Filter by HTTP method (GET, POST, PUT, HEAD...).

    -f http_method[GET]
    """

    def filter_func(log_line):
        return log_line.http_request_method == http_method

    return filter_func


def filter_backend(backend_name):
    """Filter by HAProxy backend.

    -f backend[specific_app]

    See HAProxy configuration, it can have multiple backends defined.
    """

    def filter_func(log_line):
        return log_line.backend_name == backend_name

    return filter_func


def filter_frontend(frontend_name):
    """Filter by which HAProxy frontend got the request.

    -f frontend[loadbalancer]

    See HAProxy configuration, it can have multiple frontends defined.
    """

    def filter_func(log_line):
        return log_line.frontend_name == frontend_name

    return filter_func


def filter_server(server_name):
    """Filter by downstream server.

    -f server[app01]
    """

    def filter_func(log_line):
        return log_line.server_name == server_name

    return filter_func


def filter_response_size(size):
    """Filter by how big (in bytes) the response was.

    -f response_size[50000]

    Specially useful when looking for big file downloads.
    """
    if size.startswith('+'):
        size_value = int(size[1:])
    else:
        size_value = int(size)

    def filter_func(log_line):
        bytes_read = log_line.bytes_read
        if bytes_read.startswith('+'):
            bytes_read = int(bytes_read[1:])
        else:
            bytes_read = int(bytes_read)

        return bytes_read >= size_value

    return filter_func



from datetime import datetime

import re


# Example log line, to understand the regex below (truncated to fit into
# 80 chars):
#
# Dec  9 13:01:26 localhost haproxy[28029]: 127.0.0.1:39759
# [09/Dec/2013:12:59:46.633] loadbalancer default/instance8
# 0/51536/1/48082/99627 200 83285 - - ---- 87/87/87/1/0 0/67
# {77.24.148.74} "GET /path/to/image HTTP/1.1"

HAPROXY_LINE_REGEX = re.compile(
    # Dec  9 13:01:26 localhost haproxy[28029]:
    # ignore the syslog prefix
    r'\A.*\]:\s+'
    # 127.0.0.1:39759
    r'(?P<client_ip>[a-fA-F\d+\.:]+):(?P<client_port>\d+)\s+'
    # [09/Dec/2013:12:59:46.633]
    r'\[(?P<accept_date>.+)\]\s+'
    # loadbalancer default/instance8
    r'(?P<frontend_name>.*)\s+(?P<backend_name>.*)/(?P<server_name>.*)\s+'
    # 0/51536/1/48082/99627
    r'(?P<tq>-?\d+)/(?P<tw>-?\d+)/(?P<tc>-?\d+)/'
    r'(?P<tr>-?\d+)/(?P<tt>\+?\d+)\s+'
    # 200 83285
    r'(?P<status_code>-?\d+)\s+(?P<bytes_read>\+?\d+)\s+'
    # - - ----
    r'.*\s+'  # ignored by now, should capture cookies and termination state
    # 87/87/87/1/0
    r'(?P<act>\d+)/(?P<fe>\d+)/(?P<be>\d+)/'
    r'(?P<srv>\d+)/(?P<retries>\+?\d+)\s+'
    # 0/67
    r'(?P<queue_server>\d+)/(?P<queue_backend>\d+)\s+'
    # {77.24.148.74}
    r'({(?P<request_headers>.*)}\s+{(?P<response_headers>.*)}\s+|{(?P<headers>.*)}\s+|)'
    # "GET /path/to/image HTTP/1.1"
    r'"(?P<http_request>.*)"'
    r'\Z'  # end of line
)

HTTP_REQUEST_REGEX = re.compile(
    r'(?P<method>\w+)\s+'
    r'(?P<path>(/[`´\\<>/\w:,;.#$!?=&@%_+\'*^~|()\[\]{\}-]*)+)'
    r'(\s+(?P<protocol>\w+/\d\.\d))?'
)


class Line:
    """For a precise and more detailed description of every field see:
    http://cbonte.github.io/haproxy-dconv/2.2/configuration.html#8.2.3
    """

    #: IP of the upstream server that made the connection to HAProxy.
    client_ip = None
    #: Port used by the upstream server that made the connection to HAProxy.
    client_port = None

    # raw string from log line and its python datetime version
    raw_accept_date = None
    #: datetime object with the exact date when the connection to HAProxy was
    #: made.
    accept_date = None

    #: HAProxy frontend that received the connection.
    frontend_name = None
    #: HAProxy backend that the connection was sent to.
    backend_name = None
    #: Downstream server that HAProxy send the connection to.
    server_name = None

    #: Time in milliseconds waiting the client to send the full HTTP request
    #: (``Tq`` in HAProxy documentation).
    time_wait_request = None
    #: Time in milliseconds that the request spend on HAProxy queues
    #: (``Tw`` in HAProxy documentation).
    time_wait_queues = None
    #: Time in milliseconds to connect to the final server
    #: (``Tc`` in HAProxy documentation).
    time_connect_server = None
    #: Time in milliseconds waiting the downstream server to send the full
    #: HTTP response (``Tr`` in HAProxy documentation).
    time_wait_response = None
    #: Total time in milliseconds between accepting the HTTP request and
    #: sending back the HTTP response (``Tt`` in HAProxy documentation).
    total_time = None

    #: HTTP status code returned to the client.
    status_code = None
    #: Total number of bytes send back to the client.
    bytes_read = None

    # not used by now
    captured_request_cookie = None
    captured_response_cookie = None

    # not used by now
    termination_state = None

    #: Total number of concurrent connections on the process when the
    #: session was logged (``actconn`` in HAProxy documentation).
    connections_active = None
    #: Total number of concurrent connections on the frontend when the
    #: session was logged (``feconn`` in HAProxy documentation).
    connections_frontend = None
    #: Total number of concurrent connections handled by the backend when
    #: the session was logged (``beconn`` in HAProxy documentation).
    connections_backend = None
    #: Total number of concurrent connections still active on the server
    #: when the session was logged (``srv_conn`` in HAProxy documentation).
    connections_server = None
    #: Number of connection retries experienced by this session when
    # trying to connect to the server.
    retries = None

    #: Total number of requests which were processed before this one in
    #: the server queue (``srv_queue`` in HAProxy documentation).
    queue_server = None
    #: Total number of requests which were processed before this one in
    #: the backend's global queue (``backend_queue`` in HAProxy documentation).
    queue_backend = None

    # List of headers captured in the request.
    captured_request_headers = None
    # List of headers captured in the response.
    captured_response_headers = None

    raw_http_request = None
    #: HTTP method (GET, POST...) used on this request.
    http_request_method = None
    #: Requested HTTP path.
    http_request_path = None
    #: HTTP version used on this request.
    http_request_protocol = None

    raw_line = None

    def __init__(self, line):
        self.raw_line = line

        self.is_valid = self._parse_line(line)

    @property
    def is_https(self):
        """Returns True if the log line is a SSL connection. False otherwise."""
        if ':443' in self.http_request_path:
            return True
        return False

    def is_within_time_frame(self, start, end):
        if not start:
            return True
        elif start > self.accept_date:
            return False

        if not end:
            return True
        elif end < self.accept_date:
            return False

        return True

    @property
    def ip(self):
        """Returns the IP provided on the log line, or the client_ip if absent/empty."""
        if self.captured_request_headers is not None:
            ip = self.captured_request_headers.split('|')[0]
            if ip:
                # only get the first IP, if there are more usually
                # are the intermediate servers
                return ip.split(',')[0]
        return self.client_ip

    def _parse_line(self, line):
        matches = HAPROXY_LINE_REGEX.match(line)
        if matches is None:
            return False

        self.client_ip = matches.group('client_ip')
        self.client_port = int(matches.group('client_port'))

        self.raw_accept_date = matches.group('accept_date')
        self.accept_date = self._parse_accept_date()

        self.frontend_name = matches.group('frontend_name')
        self.backend_name = matches.group('backend_name')
        self.server_name = matches.group('server_name')

        self.time_wait_request = int(matches.group('tq'))
        self.time_wait_queues = int(matches.group('tw'))
        self.time_connect_server = int(matches.group('tc'))
        self.time_wait_response = int(matches.group('tr'))
        self.total_time = matches.group('tt')

        self.status_code = matches.group('status_code')
        self.bytes_read = matches.group('bytes_read')

        self.connections_active = matches.group('act')
        self.connections_frontend = matches.group('fe')
        self.connections_backend = matches.group('be')
        self.connections_server = matches.group('srv')
        self.retries = matches.group('retries')

        self.queue_server = int(matches.group('queue_server'))
        self.queue_backend = int(matches.group('queue_backend'))

        self.captured_request_headers = matches.group('request_headers')
        self.captured_response_headers = matches.group('response_headers')
        if matches.group('headers') is not None:
            self.captured_request_headers = matches.group('headers')

        self.raw_http_request = matches.group('http_request')
        self._parse_http_request()

        return True

    def _parse_accept_date(self):
        return datetime.strptime(self.raw_accept_date, '%d/%b/%Y:%H:%M:%S.%f')

    def _parse_http_request(self):
        matches = HTTP_REQUEST_REGEX.match(self.raw_http_request)
        if matches:
            self.http_request_method = matches.group('method')
            self.http_request_path = matches.group('path')
            self.http_request_protocol = matches.group('protocol')
        else:
            self.handle_bad_http_request()

    def handle_bad_http_request(self):
        self.http_request_method = 'invalid'
        self.http_request_path = 'invalid'
        self.http_request_protocol = 'invalid'

        if self.raw_http_request != '<BADREQ>':
            print(f'Could not process HTTP request {self.raw_http_request}')


# it is not coverage covered as this is executed by the multiprocessor module,
# and setting it up on coverage just for two lines is not worth it
def parse_line(line):  # pragma: no cover
    return Line(line.strip())



from datetime import datetime
from haproxy.line import HAPROXY_LINE_REGEX
from haproxy.line import HTTP_REQUEST_REGEX

import pytest
import random


def test_default_values(line_factory, default_line_data):
    """Check that the default line with default values is parsed."""
    line = line_factory()
    matches = HAPROXY_LINE_REGEX.match(line.raw_line)
    assert matches.group('http_request') == default_line_data['http_request']


def test_client_ip_and_port(line_factory):
    """Check that the client IP and port are extracted correctly."""
    ip = '192.168.0.250'
    port = '34'
    line = line_factory(client_ip=ip, client_port=port)
    matches = HAPROXY_LINE_REGEX.match(line.raw_line)

    assert matches.group('client_ip') == ip
    assert matches.group('client_port') == port


def test_accept_date(line_factory):
    """Check that the accept date is extracted correctly."""
    accept_date = datetime.now().strftime('%d/%b/%Y:%H:%M:%S.%f')
    line = line_factory(accept_date=accept_date)
    matches = HAPROXY_LINE_REGEX.match(line.raw_line)

    assert matches.group('accept_date') == accept_date


def test_server_names(line_factory):
    """Check that the server names are extracted correctly."""
    frontend_name = 'SomeThing4'
    backend_name = 'Another1'
    server_name = 'Cloud9'
    line = line_factory(
        frontend_name=frontend_name, backend_name=backend_name, server_name=server_name
    )
    matches = HAPROXY_LINE_REGEX.match(line.raw_line)

    assert matches.group('frontend_name') == frontend_name
    assert matches.group('backend_name') == backend_name
    assert matches.group('server_name') == server_name


@pytest.mark.parametrize(
    ('tq', 'tw', 'tc', 'tr', 'tt'),
    [
        ('0', '0', '0', '0', '0'),
        ('23', '55', '3', '4', '5'),
        ('-23', '-33', '-3', '-4', '5'),
        ('23', '33', '3', '4', '+5'),
    ],
)
def test_timers(line_factory, tq, tw, tc, tr, tt):
    """Check that the timers are extracted correctly.

    Note that all timers can be negative but `tt`,
    and that `tt` is the only one that can have a positive sign.
    """
    line = line_factory(tq=tq, tw=tw, tc=tc, tr=tr, tt=tt)
    matches = HAPROXY_LINE_REGEX.match(line.raw_line)

    assert matches.group('tq') == tq
    assert matches.group('tw') == tw
    assert matches.group('tc') == tc
    assert matches.group('tr') == tr
    assert matches.group('tt') == tt


@pytest.mark.parametrize(
    ('status', 'bytes_read'), [('200', '0'), ('-301', '543'), ('200', '+543')]
)
def test_status_and_bytes(line_factory, status, bytes_read):
    """Check that the status code and bytes are extracted correctly.

    Note that `status` can be negative (for terminated requests),
    and `bytes` can be prefixed with a plus sign.
    """
    line = line_factory(status=status, bytes=bytes_read)
    matches = HAPROXY_LINE_REGEX.match(line.raw_line)

    assert matches.group('status_code') == status
    assert matches.group('bytes_read') == bytes_read


@pytest.mark.parametrize(
    ('act', 'fe', 'be', 'srv', 'retries'),
    [
        ('0', '0', '0', '0', '0'),
        ('40', '10', '11', '12', '14'),
        ('40', '10', '11', '12', '+14'),
    ],
)
def test_connections_and_retries(line_factory, act, fe, be, srv, retries):
    """Check that the connections and retries are extracted correctly.

    Note that `retries` might have a plus sign prefixed.
    """
    line = line_factory(act=act, fe=fe, be=be, srv=srv, retries=retries)
    matches = HAPROXY_LINE_REGEX.match(line.raw_line)

    assert matches.group('act') == act
    assert matches.group('fe') == fe
    assert matches.group('be') == be
    assert matches.group('srv') == srv
    assert matches.group('retries') == retries


@pytest.mark.parametrize(('server', 'backend'), [('0', '0'), ('200', '200')])
def test_queues(line_factory, server, backend):
    """Check that the server and backend queues are extracted correctly."""
    line = line_factory(queue_server=server, queue_backend=backend)
    matches = HAPROXY_LINE_REGEX.match(line.raw_line)

    assert matches.group('queue_server') == server
    assert matches.group('queue_backend') == backend


@pytest.mark.parametrize(
    ('request_header', 'response_header'),
    [
        ('', ''),
        ('something', None),
        ('something here', 'and there'),
        ('multiple | request | headers', 'and | multiple | response ones'),
    ],
)
def test_captured_headers(line_factory, request_header, response_header):
    """Check that captured headers are extracted correctly."""
    if response_header:
        headers = f' {{{request_header}}} {{{response_header}}}'
    else:
        headers = f' {{{request_header}}}'
    line = line_factory(headers=headers)
    matches = HAPROXY_LINE_REGEX.match(line.raw_line)

    if response_header:
        assert matches.group('request_headers') == request_header
        assert matches.group('response_headers') == response_header
    else:
        assert matches.group('headers') == request_header
        assert matches.group('request_headers') is None
        assert matches.group('response_headers') is None


def test_http_request(line_factory):
    """Check that the HTTP request is extracted correctly."""
    http_request = 'something in the air'
    line = line_factory(http_request=http_request)
    matches = HAPROXY_LINE_REGEX.match(line.raw_line)

    assert matches.group('http_request') == http_request


@pytest.mark.parametrize(
    'path',
    [
        '/path/to/image',
        '/path/with/port:80',  # with port
        '/path/with/example.com',  # with domain
        '/path/to/article#section',  # with anchor
        '/article?hello=world&goodbye=lennin',  # with parameters
        '/article-with-dashes_and_underscores',  # dashes and underscores
        '/redirect_to?http://example.com',  # double slashes
        '/@@funny',  # at sign
        '/something%20encoded',  # percent sign
        '/++adding++is+always+fun',  # plus sign
        '/here_or|here',  # vertical bar
        '/here~~~e',  # tilde sign
        '/here_*or',  # asterisk sign
        '/something;or-not',  # colon
        '/something-important!probably',  # exclamation mark
        '/something$important',  # dollar sign
        "/there's-one's-way-or-another's"  # single quote sign
        '/there?la=as,is',  # comma
        '/here_or(here)',  # parenthesis
        '/here_or[here]',  # square brackets
        '/georg}von{grote/\\',  # curly brackets
        '/here_or<',  # less than
        '/here_or>',  # more than
        '/georg-von-grote/\\',  # back slash
        '/georg`von´grote/\\',  # diacritics
        '/georg`von^grote/\\',  # caret
    ],
)
def test_http_request_regex(path):
    """Test that the method/path/protocol are extracted properly from the HTTP request."""
    verbs = ('GET', 'POST', 'DELETE', 'PATCH', 'PUT')
    protocols = (
        'HTTP/1.0',
        'HTTP/1.1',
        'HTTP/2.0',
    )
    method = random.choice(verbs)
    protocol = random.choice(protocols)
    matches = HTTP_REQUEST_REGEX.match(f'{method} {path} {protocol}')
    assert matches.group('method') == method
    assert matches.group('path') == path
    assert matches.group('protocol') == protocol



from datetime import datetime
from haproxy.line import parse_line
from haproxy.utils import date_str_to_datetime
from haproxy.utils import delta_str_to_timedelta
from multiprocessing import Pool


class Log:
    def __init__(self, logfile=None, start=None, delta=None, show_invalid=False):
        self.logfile = logfile
        self.show_invalid = show_invalid
        self.start = None
        self.end = None

        if start:
            self.start = date_str_to_datetime(start)

        if delta:
            delta = delta_str_to_timedelta(delta)

            if isinstance(self.start, datetime):
                self.end = self.start + delta

        self.invalid_lines = 0
        self.valid_lines = 0

    def __iter__(self):
        start = datetime.now()
        with open(self.logfile) as logfile, Pool() as pool:
            for index, line in enumerate(pool.imap(parse_line, logfile)):
                if line.is_valid:
                    self.valid_lines += 1
                    if line.is_within_time_frame(self.start, self.end):
                        yield line
                else:
                    if self.show_invalid:
                        print(line.raw_line)
                    self.invalid_lines += 1

                if index % 10000 == 0 and index > 0:  # pragma: no cover
                    print('.', end='', flush=True)

        end = datetime.now()
        print(f'\nIt took {end - start}')

    @property
    def total_lines(self):
        return self.valid_lines + self.invalid_lines



from datetime import datetime
from datetime import timedelta

import re


DELTA_REGEX = re.compile(r'\A(?P<value>\d+)(?P<time_unit>[smhd])\Z')

START_REGEX = re.compile(
    r'(?P<day>\d+)/(?P<month>\w+)/(?P<year>\d+)'
    r'(:(?P<hour>\d+)|)(:(?P<minute>\d+)|)(:(?P<second>\d+)|)'
)

DELTA_KEYS = {'s': 'seconds', 'm': 'minutes', 'h': 'hours', 'd': 'days'}


def date_str_to_datetime(date):
    """Convert a string to a datetime object.

    The format is `day/month/year[[[:hour]:minute]:second]` being:
    - day a number
    - month a three letter representation of the month (i.e. Dec, Jan, etc)
    - year as a 4 digits value
    - hour/minute/second as 2 digits value, each of them being optional
    """
    matches = START_REGEX.match(date)
    data = matches.group('day'), matches.group('month'), matches.group('year')
    raw_date_input = f'{data[0]}/{data[1]}/{data[2]}'
    date_format = '%d/%b/%Y'
    for variable, percent in (('hour', ':%H'), ('minute', ':%M'), ('second', ':%S')):
        match = matches.group(variable)
        if match:
            date_format += percent
            raw_date_input = f'{raw_date_input}:{match}'

    return datetime.strptime(raw_date_input, date_format)


def delta_str_to_timedelta(delta):
    """Convert a string to a timedelta representation.

    Format is NUMBER followed by one of the following letters: `s`, `m`, `h`, `d`.
    Each of them meaning, second, minute, hour and day.
    """
    matches = DELTA_REGEX.match(delta)
    value = int(matches.group('value'))
    time_unit = matches.group('time_unit')
    key = DELTA_KEYS[time_unit]
    return timedelta(**{key: value})


def validate_arg_date(start):
    """Check that date argument is valid."""
    try:
        date_str_to_datetime(start)
    except (AttributeError, ValueError):
        raise ValueError('--start argument is not valid')


def validate_arg_delta(delta):
    """Check that the delta argument is valid."""
    try:
        delta_str_to_timedelta(delta)
    except (AttributeError, ValueError):
        raise ValueError('--delta argument is not valid')


def list_filters():
    """Return the information of existing filters.

    Data returned:
    - their names as the user is expected to use them from the command line
    - the object itself
    - its description
    """
    from haproxy import filters

    data = {}
    for full_name in dir(filters):
        if not full_name.startswith('filter_'):
            continue
        name = full_name[7:]
        obj = getattr(filters, full_name)

        description = _strip_description(obj.__doc__)
        data[name] = {'obj': obj, 'description': f'{name}:\n\t{description}'}
    return data


def list_commands():
    """Return the information of existing commands.

    Data returned:
    - their names as the user is expected to use them from the command line
    - the object itself
    - its description
    """
    from haproxy import commands

    data = {}
    for cmd in dir(commands):
        if cmd.endswith('Mixin'):
            continue
        klass = getattr(commands, cmd)
        try:
            name = klass.command_line_name()
        except AttributeError:
            continue

        description = _strip_description(klass.__doc__)
        data[name] = {'klass': klass, 'description': f'{name}:\n\t{description}'}
    return data


def _strip_description(raw_text):
    if not raw_text:
        return ''
    text = '\n\t'.join([line.strip() for line in raw_text.split('\n') if line.strip()])
    return text


VALID_COMMANDS = list_commands()
VALID_FILTERS = list_filters()
